# game.py
# ---------
# COSC4550/COSC5550
# Artificial Intelligence
# University of Wyoming
# ---------
# Most code was part of the Pacman AI projects by John DeNero and Dan Klein
# Code has been modified for the Cleaning Agents AI Challenge by Joost Huizinga
# Licensing Information: Please do not distribute or publish solutions to this
# project. You are free to use and extend these projects for educational
# purposes. The Pacman AI projects were developed at UC Berkeley, primarily by
# John DeNero (denero@cs.berkeley.edu) and Dan Klein (klein@cs.berkeley.edu).
# For more info, see http://inst.eecs.berkeley.edu/~cs188/sp09/pacman.html

from util import *
import time, os
import traceback
  
#######################
# Parts worth reading #
#######################

class Agent:
  """
  An agent must define a getAction method, but may also define the
  following methods which will be called if they exist:

  def registerInitialState(self, state): # inspects the starting state
  """
  def __init__(self, index=0):
    self.index = index

  def getAction(self, state):
    """
    The Agent will receive a GameState (from either {pacman, capture, sonar}.py) and
    must return an action from Directions.{North, South, East, West, Stop}
    """
    raiseNotDefined()

class Directions:
  NORTH = 'North'
  SOUTH = 'South'
  EAST = 'East'
  WEST = 'West'
  STOP = 'Stop'
  NONE = 'None'

  LEFT =       {NORTH: WEST,
                 SOUTH: EAST,
                 EAST:  NORTH,
                 WEST:  SOUTH,
                 STOP:  STOP,
                 NONE: NONE}

  RIGHT =      dict([(y,x) for x, y in LEFT.items()])

  REVERSE = {NORTH: SOUTH,
             SOUTH: NORTH,
             EAST: WEST,
             WEST: EAST,
             STOP: STOP,
             NONE: NONE}


class RobotActions:
  TURN_LEFT = 'Left'
  TURN_RIGHT = 'Right'
  TURN_AND_MOVE_LEFT = 'MoveLeft'
  TURN_AND_MOVE_RIGHT = 'MoveRight'
  FORWARD = 'Forward'
  DO_NOTHING = 'Stop'
  STOP = 'Stop' #inlcuded for lagecy reasons, use DO_NOTHING instead

class Configuration:
  """
  A Configuration holds the (x,y) coordinate of a character, along with its
  traveling direction.

  The convention for positions, like a graph, is that (0,0) is the lower left corner, x increases
  horizontally and y increases vertically.  Therefore, north is the direction of increasing y, or (0,1).
  """

  def __init__(self, pos, direction):
    self.pos = pos
    self.direction = direction

  def getPosition(self):
    return (self.pos)

  def getDirection(self):
    return self.direction

  def isInteger(self):
    x,y = self.pos
    return x == int(x) and y == int(y)

  def __eq__(self, other):
    if other == None: return False
    return (self.pos == other.pos and self.direction == other.direction)

  def __hash__(self):
    x = hash(self.pos)
    y = hash(self.direction)
    return hash(x + 13 * y)

  def __str__(self):
    return "(x,y)="+str(self.pos)+", "+str(self.direction)

  def generateSuccessor(self, newdirection, vector):
    """
    Generates a new configuration reached by translating the current
    configuration by the action vector.  This is a low-level call and does
    not attempt to respect the legality of the movement.

    Actions are movement vectors.
    """

    x, y= self.pos
    dx, dy = vector
    return Configuration((x + dx, y+dy), newdirection)

class AgentState:
  """
  AgentStates hold the state of an agent (configuration, speed, sensor information, etc).
  """

  def __init__( self, startConfiguration):
    self.start = startConfiguration
    self.configuration = startConfiguration
    self.leftWallSensor = 0
    self.rightWallSensor = 0
    self.frontWallSensor = 0
    self.backWallSensor = 0
    self.frontDustSensor = 0
    self.leftDustSensor = 0
    self.rightDustSensor = 0
    self.dustUtility = 0
    self.draft = Directions.NONE

  def __str__( self ):
      return "Agent: " + str( self.configuration )


  def __eq__( self, other ):
    if other == None:
      return False
    return self.configuration == other.configuration

  def __hash__(self):
    return hash(self.configuration)

  def copy( self ):
    state = AgentState( self.start )
    state.configuration = self.configuration
    state.draft = self.draft
    state.dustUtility = self.dustUtility
    return state

  def getPosition(self):
    if self.configuration == None: return None
    return self.configuration.getPosition()

  def getDirection(self):
    return self.configuration.getDirection()







class Grid:
  """
  A 2-dimensional array of objects backed by a list of lists.  Data is accessed
  via grid[x][y] where (x,y) are positions on a Pacman map with x horizontal,
  y vertical and the origin (0,0) in the bottom left corner.

  The __str__ method constructs an output that is oriented like a pacman board.
  """
  def __init__(self, width, height, initialValue=False, bitRepresentation=None):
    if initialValue not in [False, True]: raise Exception('Grids can only contain booleans')
    self.CELLS_PER_INT = 30
    self.default = initialValue
    self.width = width
    self.height = height
    self.data = [[initialValue for y in range(height)] for x in range(width)]
    if bitRepresentation:
      self._unpackBits(bitRepresentation)

  def __getitem__(self, i):
    return self.data[i]

  def __setitem__(self, key, item):
    self.data[key] = item

  def __str__(self):
    out = [[str(self.data[x][y])[0] for x in range(self.width)] for y in range(self.height)]
    out.reverse()
    return '\n'.join([''.join(x) for x in out])

  def __eq__(self, other):
    if other == None: return False
    return self.data == other.data

  def __hash__(self):
    # return hash(str(self))
    base = 1
    h = 0
    for l in self.data:
      for i in l:
        if i:
          h += base
        base *= 2
    return hash(h)

  def copy(self):
    g = Grid(self.width, self.height)
    g.data = [x[:] for x in self.data]
    return g

  def deepCopy(self):
    return self.copy()

  def shallowCopy(self):
    g = Grid(self.width, self.height)
    g.data = self.data
    return g

  def count(self, item =True ):
    return sum([x.count(item) for x in self.data])

  def asList(self, key = True):
    list = []
    for x in range(self.width):
      for y in range(self.height):
        if self[x][y] == key: list.append( (x,y) )
    return list

  def safeGet(self, x, y):
    if(x>= self.width or x<0 or y>= self.height or y<0):
      return self.default
    return self[x][y]

  def safeSet(self, x, y, value):
    if(x>= self.width or x<0 or y>= self.height or y<0):
      return 
    self[x][y] = value

  def packBits(self):
    """
    Returns an efficient int list representation

    (width, height, bitPackedInts...)
    """
    bits = [self.width, self.height]
    currentInt = 0
    for i in range(self.height * self.width):
      bit = self.CELLS_PER_INT - (i % self.CELLS_PER_INT) - 1
      x, y = self._cellIndexToPosition(i)
      if self[x][y]:
        currentInt += 2 ** bit
      if (i + 1) % self.CELLS_PER_INT == 0:
        bits.append(currentInt)
        currentInt = 0
    bits.append(currentInt)
    return tuple(bits)

  def _cellIndexToPosition(self, index):
    x = index / self.height
    y = index % self.height
    return x, y

  def _unpackBits(self, bits):
    """
    Fills in data from a bit-level representation
    """
    cell = 0
    for packed in bits:
      for bit in self._unpackInt(packed, self.CELLS_PER_INT):
        if cell == self.width * self.height: break
        x, y = self._cellIndexToPosition(cell)
        self[x][y] = bit
        cell += 1

  def _unpackInt(self, packed, size):
    bools = []
    if packed < 0: raise ValueError, "must be a positive integer"
    for i in range(size):
      n = 2 ** (self.CELLS_PER_INT - i - 1)
      if packed >= n:
        bools.append(True)
        packed -= n
      else:
        bools.append(False)
    return bools

def reconstituteGrid(bitRep):
  if type(bitRep) is not type((1,2)):
    return bitRep
  width, height = bitRep[:2]
  return Grid(width, height, bitRepresentation= bitRep[2:])



####################################
# Parts you shouldn't have to read #
####################################





class Actions:
  """
  A collection of static methods for manipulating move actions.
  """
  # Directions
  _directions = {Directions.NORTH: (0, 1),
                 Directions.SOUTH: (0, -1),
                 Directions.EAST:  (1, 0),
                 Directions.WEST:  (-1, 0),
                 Directions.STOP:  (0, 0)}

  _directionsAsList = _directions.items()

  TOLERANCE = .001

  def reverseDirection(action):
    if action == Directions.NORTH:
      return Directions.SOUTH
    if action == Directions.SOUTH:
      return Directions.NORTH
    if action == Directions.EAST:
      return Directions.WEST
    if action == Directions.WEST:
      return Directions.EAST
    return action
  reverseDirection = staticmethod(reverseDirection)

  def vectorToDirection(vector):
    dx, dy = vector
    if dy > 0:
      return Directions.NORTH
    if dy < 0:
      return Directions.SOUTH
    if dx < 0:
      return Directions.WEST
    if dx > 0:
      return Directions.EAST
    return Directions.STOP
  vectorToDirection = staticmethod(vectorToDirection)

  def directionToVector(direction, speed = 1.0):
    dx, dy =  Actions._directions[direction]
    return (dx * speed, dy * speed)
  directionToVector = staticmethod(directionToVector)

  def getPossibleActions(config, walls):
    possible = []
    x, y = config.pos
    x_int, y_int = int(x + 0.5), int(y + 0.5)

    # In between grid points, all agents must continue straight
    if (abs(x - x_int) + abs(y - y_int)  > Actions.TOLERANCE):
      return [config.getDirection()]

    for dir, vec in Actions._directionsAsList:
      dx, dy = vec
      next_y = y_int + dy
      next_x = x_int + dx
      if not walls[next_x][next_y]: possible.append(dir)

    return possible

  getPossibleActions = staticmethod(getPossibleActions)

  def getLegalNeighbors(position, walls):
    x,y = position
    x_int, y_int = int(x + 0.5), int(y + 0.5)
    neighbors = []
    for dir, vec in Actions._directionsAsList:
      dx, dy = vec
      next_x = x_int + dx
      if next_x < 0 or next_x == walls.width: continue
      next_y = y_int + dy
      if next_y < 0 or next_y == walls.height: continue
      if not walls[next_x][next_y]: neighbors.append((next_x, next_y))
    return neighbors
  getLegalNeighbors = staticmethod(getLegalNeighbors)

  def getSuccessor(position, action):
    dx, dy = Actions.directionToVector(action)
    x, y = position
    return (x + dx, y + dy)
  getSuccessor = staticmethod(getSuccessor)







class GameStateData:
  """

  """
  def __init__( self, prevState = None ):
    """
    Generates a new data packet by copying information from its predecessor.
    """
    if prevState != None:
      self.dust = prevState.dust.shallowCopy()
      self.agentStates = self.copyAgentStates( prevState.agentStates )
      self.layout = prevState.layout
      self.score = prevState.score
      self.timer = prevState.timer
      self.totalDust = prevState.totalDust
      self.dustPos = prevState.dustPos
      self.prevDustPos = prevState.prevDustPos
      self.newDustPos = prevState.newDustPos
      self.utility = prevState.utility
      self.dustUtility = prevState.dustUtility
      self.energyUtility = prevState.energyUtility
    self._dustRemoved = None
    self._agentMoved = None
    self._finished = False
    self.scoreChange = 0

  def deepCopy( self ):
    state = GameStateData( self )
    state.dust = self.dust.deepCopy()
    state.layout = self.layout.deepCopy()
    state._agentMoved = self._agentMoved
    state._dustRemoved = self._dustRemoved
    state.timer = self.timer
    state.totalDust = self.totalDust
    state.dustPos = self.dustPos
    state.prevDustPos = self.prevDustPos
    state.newDustPos = self.newDustPos
    state.utility = self.utility
    state.dustUtility = self.dustUtility
    state.energyUtility = self.energyUtility
    return state

  def copyAgentStates( self, agentStates ):
    copiedStates = []
    for agentState in agentStates:
      copiedStates.append( agentState.copy() )
    return copiedStates

  def __eq__( self, other ):
    """
    Allows two states to be compared.
    """
    if other == None: return False
    # TODO Check for type of other
    if not self.agentStates == other.agentStates: return False
    if not self.dust == other.dust: return False
    if not self.score == other.score: return False
    if not self.timer == other.timer: return False
    if not self.totalDust == other.totalDust: return False
    return True

  def __hash__( self ):
    """
    Allows states to be keys of dictionaries.
    """
    for i, state in enumerate( self.agentStates ):
      try:
        int(hash(state))
      except TypeError, e:
        print e
        #hash(state)
    return int((hash(tuple(self.agentStates)) + 13*hash(self.dust) + 113*hash(self.timer) + 7 * hash(self.score)) % 1048575 )

  def __str__( self ):
    width, height = self.layout.width, self.layout.height
    map = Grid(width, height)
    if type(self.dust) == type((1,2)):
      self.dust = reconstituteGrid(self.dust)
    for x in range(width):
      for y in range(height):
        dust, walls = self.dust, self.layout.walls
        map[x][y] = self._foodWallStr(dust[x][y], walls[x][y])

    for agentState in self.agentStates:
      if agentState == None: continue
      if agentState.configuration == None: continue
      x,y = [int( i ) for i in nearestPoint( agentState.configuration.pos )]
      agent_dir = agentState.configuration.direction
      map[x][y] = self._agentStr( agent_dir )

    return str(map) + ("\nScore: %d\n" % self.score)

  def _foodWallStr( self, hasDust, hasWall ):
    if hasDust:
      return '.'
    elif hasWall:
      return '%'
    else:
      return ' '

  def _agentStr( self, dir ):
    if dir == Directions.NORTH:
      return 'v'
    if dir == Directions.SOUTH:
      return '^'
    if dir == Directions.WEST:
      return '>'
    return '<'

  def initialize( self, layout):
    """
    Creates an initial game state from a layout array (see layout.py).
    """
    self.dust = layout.dust.copy()
    self.layout = layout
    self.score = 0
    self.scoreChange = 0
    self.timer = 200
    self.totalDust = self.dust.count()
    self.dustPos = []
    self.prevDustPos = []
    self.newDustPos = []
    self.utility = True
    base = random.randint(1, 6)
    base = base*base
    self.dustUtility = base + 1
#    self.dustUtility = 40
    self.energyUtility = 1

    for x in range(layout.width):
      for y in range(layout.height):
        if self.dust[x][y]:
          self.dustPos.append((x,y))
#          self.prevDustPos.append((x,y))

    self.agentStates = []
    for pos in layout.agentPositions:
      self.agentStates.append( AgentState( Configuration( pos, Directions.EAST)) )
    self.agentStates[0].dustUtility = self.dustUtility
    self._dustRemoved = [False for a in self.agentStates]











class Game:
  """
  The Game manages the control flow, soliciting actions from agents.
  """

  def __init__( self, agents, display, rules, startingIndex=0, muteAgents=False, catchExceptions=False, useDraft=False):
    self.agentCrashed = False
    self.agents = agents
    self.display = display
    self.rules = rules
    self.startingIndex = startingIndex
    self.gameOver = False
    self.muteAgents = muteAgents
    self.catchExceptions = catchExceptions
    self.moveHistory = []
    self.totalAgentTimes = [0 for agent in agents]
    self.totalAgentTimeWarnings = [0 for agent in agents]
    self.agentTimeout = False
    self.useDraft = useDraft
    import cStringIO
    self.agentOutput = [cStringIO.StringIO() for agent in agents]

  def getProgress(self):
    if self.gameOver:
      return 1.0
    else:
      return self.rules.getProgress(self)

  def _agentCrash( self, agentIndex, quiet=False):
    "Helper method for handling agent crashes"
    if not quiet: traceback.print_exc()
    self.gameOver = True
    self.agentCrashed = True
    self.rules.agentCrash(self, agentIndex)

  def observationFunction(self, state):
    walls = state.getWalls()
    dust = state.getDust()
    agentState = state.data.agentStates[0]
    direct = agentState.getDirection()
    pos = agentState.getPosition()


    i = 1
    north = False
    east = False
    south = False
    west = False
    northDust = 0.0
    southDust = 0.0
    westDust = 0.0
    eastDust = 0.0

    while(not(north) or not(east) or not(south) or not(west)):
      if(not(north)):
        north = walls[pos[0]][pos[1]+i]
        inorth = i
        if(i<=2 and dust[pos[0]][pos[1]+i]): northDust += 1.0/float(i) 
      if(not(east)):
        east = walls[pos[0]+i][pos[1]]
        ieast = i
        if(i<=2 and dust[pos[0]+i][pos[1]]): eastDust += 1.0/float(i) 
      if(not(south)):
        south = walls[pos[0]][pos[1]-i]
        isouth = i
        if(i<=2 and dust[pos[0]][pos[1]-i]): southDust += 1.0/float(i) 
      if(not(west)):
        west = walls[pos[0]-i][pos[1]]
        iwest = i
        if(i<=2 and dust[pos[0]-i][pos[1]]): westDust += 1.0/float(i) 
      i = i + 1
 
   

    if(direct == Directions.NORTH):
      agentState.leftWallSensor = iwest
      agentState.rightWallSensor = ieast
      agentState.frontWallSensor = inorth
      agentState.backWallSensor = isouth
      agentState.frontDustSensor = northDust 
      agentState.leftDustSensor = westDust
      agentState.rightDustSensor = eastDust
    if(direct == Directions.EAST):
      agentState.leftWallSensor = inorth 
      agentState.rightWallSensor = isouth
      agentState.frontWallSensor = ieast
      agentState.backWallSensor = iwest
      agentState.frontDustSensor = eastDust 
      agentState.leftDustSensor = northDust
      agentState.rightDustSensor = southDust
    if(direct == Directions.SOUTH):
      agentState.leftWallSensor = ieast
      agentState.rightWallSensor = iwest
      agentState.frontWallSensor = isouth
      agentState.backWallSensor = inorth
      agentState.frontDustSensor = southDust 
      agentState.leftDustSensor = eastDust
      agentState.rightDustSensor = westDust
    if(direct == Directions.WEST):
      agentState.leftWallSensor = isouth 
      agentState.rightWallSensor = inorth
      agentState.frontWallSensor = iwest
      agentState.backWallSensor = ieast
      agentState.frontDustSensor = westDust 
      agentState.leftDustSensor = southDust
      agentState.rightDustSensor = northDust

    return agentState

  def run( self ):
    """
    Main control loop for game play.
    """
    self.display.initialize(self.state.data)
    self.numMoves = 0

    agentIndex = self.startingIndex
    numAgents = len( self.agents )

    numAgents+=0

    while not self.gameOver:
      # Fetch the next agent
      
      move_time = 0
      skip_action = False


      agent = self.agents[agentIndex]
        # Generate an observation of the state
      observation = self.observationFunction(self.state.deepCopy())

        # Solicit an action
      action = agent.getAction(observation)

        # Execute the action
      self.moveHistory.append( (agentIndex, action) )
      self.state = self.state.generateSuccessor( agentIndex, action )



      # Change the display
      self.display.update( self.state.data )

      # Allow for game specific conditions (winning, losing, etc.)
      self.rules.process(self.state, self)

      # Track progress
      if agentIndex +1 == numAgents: 
        self.numMoves += 1
        gust = (0,0)
        if self.useDraft and random.random() < .2:
          
          rand = random.random()
          if rand <.25:
            gust = (1, 0)
            self.state.data.agentStates[0].draft = Directions.NORTH
          elif rand <.5:
            gust = (0, 1)
            self.state.data.agentStates[0].draft = Directions.EAST
          elif rand <.75:
            gust = (-1, 0)
            self.state.data.agentStates[0].draft = Directions.SOUTH
          else:
            gust = (0, -1)
            self.state.data.agentStates[0].draft = Directions.WEST
        else:
          self.state.data.agentStates[0].draft = Directions.NONE
        self.state = self.state.moveDust(gust)


      # Next agent
      agentIndex = ( agentIndex + 1 ) % numAgents

    self.display.finish()

