# layout.py
# ---------
# COSC4550/COSC5550
# Artificial Intelligence
# University of Wyoming
# ---------
# Most code was part of the Pacman AI projects by John DeNero and Dan Klein
# Code has been modified for the Cleaning Agents AI Challenge by Joost Huizinga
# Licensing Information: Please do not distribute or publish solutions to this
# project. You are free to use and extend these projects for educational
# purposes. The Pacman AI projects were developed at UC Berkeley, primarily by
# John DeNero (denero@cs.berkeley.edu) and Dan Klein (klein@cs.berkeley.edu).
# For more info, see http://inst.eecs.berkeley.edu/~cs188/sp09/pacman.html

from util import manhattanDistance
from game import Grid
import os
import random
import copy

VISIBILITY_MATRIX_CACHE = {}

class Layout:
  """
  A Layout manages the static information about the game board.
  """
  
  def __init__(self):
    self.width = 0
    self.height = 0
    self.walls = None
    self.dust = None
    self.agentPositions = []
    self.layoutText = None
    
  def isWall(self, pos):
    x, col = pos
    return self.walls[x][col]
  
  def getRandomLegalPosition(self):
    x = random.choice(range(self.width))
    y = random.choice(range(self.height))
    while self.isWall( (x, y) ):
      x = random.choice(range(self.width))
      y = random.choice(range(self.height))
    return (x,y)


  def __str__(self):
    return "\n".join(self.layoutText)
    
  def deepCopy(self):
    newLayout = Layout()
    newLayout.width = self.width
    newLayout.height = self.height
    newLayout.walls = self.walls.deepCopy()
    newLayout.dust = self.dust.deepCopy()
    newLayout.agentPositions = copy.deepcopy(self.agentPositions)
    newLayout.layoutText = copy.deepcopy(self.layoutText)
    return newLayout

  def recursiveRandom(self, x, y, split):
    if(x <= 0 or y <=0 or (x+1)>=self.width or (y+1) >=self.height): return

    if(random.random()<0.20 and self.walls[x][y]):
      self.dust[x][y] = True

    self.walls[x][y] = False


    if(random.random()<0.02 and split > 0):
      self.recursiveRandom(x, y, split -1)
      self.recursiveRandom(x, y, split -1)
    elif split > 0:
      direction = random.random()
      if(direction<0.25):
        self.recursiveRandom(x-1, y, split)
      elif(direction<0.5):
        self.recursiveRandom(x+1, y, split)
      elif(direction<0.75):
        self.recursiveRandom(x, y-1, split)
      else:
        self.recursiveRandom(x, y+1, split)
        
      

  def randomLayout(self, width, height):
    self.width = width
    self.height = height
    self.walls = Grid(self.width, self.height, True)
    self.dust = Grid(self.width, self.height, False)
    self.layoutText = ""
    self.agentPositions = []
    self.agentPositions.append(  (10, 10)  )
    self.recursiveRandom(10,10, 6)
    
    
  def processLayoutText(self, layoutText):
    """
    Coordinates are flipped from the input format to the (x,y) convention here
    
    The shape of the maze.  Each character  
    represents a different type of object.   
     % - Wall                               
     . - Dust
     P - Agent
    Other characters are ignored.
    """
    self.width = len(layoutText[0])
    self.height = len(layoutText)
    self.walls = Grid(self.width, self.height, False)
    self.dust = Grid(self.width, self.height, False)
    self.layoutText = layoutText

    maxY = self.height - 1
    for y in range(self.height):       
      for x in range(self.width):
        layoutChar = layoutText[maxY - y][x]  
        self.processLayoutChar(x, y, layoutChar)
  
  def processLayoutChar(self, x, y, layoutChar):
    if layoutChar == '%':      
      self.walls[x][y] = True
    elif layoutChar == '.':
      self.dust[x][y] = True   
    elif layoutChar == 'P':    
      self.agentPositions.append(  (x, y)  )

def getLayout(name, back = 2):
  layout = Layout()
  layout.randomLayout(20, 20)
  while layout.dust.count()<10: layout.randomLayout(20, 20)

#  if name.endswith('.lay'):
#    layout = tryToLoad('layouts/' + name)
#    if layout == None: layout = tryToLoad(name)
#  else:
#    layout = tryToLoad('layouts/' + name + '.lay')
#    if layout == None: layout = tryToLoad(name + '.lay')
#  if layout == None and back >= 0:
#    curdir = os.path.abspath('.')
#    os.chdir('..')
#    layout = getLayout(name, back -1)
#    os.chdir(curdir)
  return layout

def tryToLoad(fullname):
  if(not os.path.exists(fullname)): return None
  f = open(fullname)
  try: 
    layout = Layout()
    layout.processLayoutText([line.strip() for line in f])
    return layout
  finally: f.close()
